//  A small programming language compiler for Visual C++ on Windows x86(32bit).
//  Full source code of the compiler included in this file.
//
//  COPYRIGHT : Public domain
//
//  HOW TO BUILD : > cl /EHsc compiler_c++_ver.cpp
//
//  USAGE : Run it in console, it reads the program text from standard-input,
//      compiles and executes on-memory. No command line parameters. When
//      inputting from keyboard, don't forget to input CONTROL-Z to end input.
//
//  GRAMMAR : Written in BNF below.
//
//
//  program ::= statement [new_line program]
//
//  statement ::=
//      array_definition | label_definition | goto_statement | if_statement |
//      assign_statement | print_statement | null_statement
//
//  array_definition ::= 'array' identifier bracketed_number
//
//  label_definition ::= identifier ':'
//
//  if_statement ::= 'if' condition 'then' statement
//
//  goto_statement ::= 'goto' identifier
//
//  assign_statement ::= variable '=' expression
//
//  print_statement ::= expression | text
//
//  null_statement ::=
//
//  condition ::= expression relational_operator expression
//
//  expression ::= term [additive_operator expression]
//
//  term ::= [unary_operator] factor [multiplicative_operator term]
//
//  factor ::= number | variable | parenthesised_expression
//
//  parenthesised_expression ::= '(' expression ')'
//
//  bracketed_expression ::= '[' expression ']'
//
//  bracketed_number ::= '[' number ']'
//
//  variable ::= identifier [bracketed_expression]
//
//  identifier ::= alphabet [identifier_subsequence]
//
//  identifier_subsequence ::= identifier_letter [identifier_subsequence]
//
//  number ::= digit [number]
//
//  text ::= '"' text_letters '"'
//
//  new_line ::= '\n' | ';'
//
//  additive_operator ::= '+' | '-'
//
//  multiplicative_operator ::= '*' | '/' | '%'
//
//  unary_operator ::= '+' | '-'
//
//  relational_operator ::= '=' | '!=' | '<' | '<=' | '>' | '>='
//
//  identifier_letter ::= digit | alphabet | '_'
//
//  digit ::= '0' .. '9'
//
//  alphabet ::= 'A' .. 'Z' | 'a' .. 'z'
//
//  text_letters ::= any 0 or more characters, but '\' '"' sequence represents a
//      quotation mark, '\' 'a' is a bell, '\' 'n' is a line feed, '\' 't' is a
//      horizontal tab and '\' '\' sequence is a backslash itself.

#define _SCL_SECURE_NO_WARNINGS
#include <iostream>
#include <map>
#include <string>
#include <algorithm>
#include <time.h>
#include <windows.h>

using std::cout;using std::cin;using std::flush;using std::endl; using std::map;
using std::string; using std::copy; using std::find_if; using std::count;

class test {
public:
    test();
    ~test();
    bool contains(const string&, int*&);
    operator bool() { return true; }

    test *parent;
    map<string, int*> identifiers;
    int *data_ptr;
    char *instruction_ptr, **jump_ptr;
    string *text_ptr;
    string::iterator code_itr;
};

int data[4096*4096];
char *instructions = static_cast<char*>(
    VirtualAlloc(0, 4096*4096, MEM_COMMIT, PAGE_READWRITE));
char *jump_stack[4096];
string texts[4096], code;
test *ct; // 'ct' is short for 'current test'.

bool entry_code(), exit_code(), expression(bool (*op)() = 0), program();

test::test() : parent(ct), identifiers(),
    data_ptr(       ct ? ct->data_ptr         : data),
    instruction_ptr(ct ? ct->instruction_ptr  : instructions),
    jump_ptr(       ct ? ct->jump_ptr         : jump_stack),
    text_ptr(       ct ? ct->text_ptr         : texts),
    code_itr(       ct ? ct->code_itr         : code.begin()) {
    ct = this;
}

test::~test() { ct = parent; }

bool test::contains(const string &s, int *&address) {
    if (parent && parent->contains(s, address)) return true;
    map<string, int*>::iterator it = identifiers.find(s);
    return it != identifiers.end() && (address = it->second);
}

bool ok() {
    ct->identifiers.insert(ct->parent->identifiers.begin(),
                           ct->parent->identifiers.end());
    test* cpp = ct->parent->parent;
    *ct->parent = *ct;
    ct->parent->parent = cpp;
    return true;
}

void compile() {
    char c;
    while (cin.get(c)) code.push_back(c);
    test t; // Prepares and sustains 'ct' : current test (of compilation).
    entry_code();
    program(); // Parses and generates code.
    if (t.code_itr != code.end()) {
        cout << "Syntax error at line " <<
            count(code.begin(), t.code_itr, '\n') + 1 << '.' << endl;
        exit(1);
    }
    exit_code();
    DWORD old;
    VirtualProtect(instructions, 4096*4096, PAGE_EXECUTE, &old);
}

void run() {
    double c = clock();
    reinterpret_cast<void(*)()>(instructions)(); // Calls compiled code.
    cout<< "\nCalculated in " << (clock() - c) / CLK_TCK << " seconds." << endl;
}

int main() {
    compile();
    run();
}

template<size_t S> bool generate(const unsigned char (&code)[S]) {
    ct->instruction_ptr = copy(code, code + S, ct->instruction_ptr);
    return true;
}

template<size_t S, typename T>
bool generate(const unsigned char (&code)[S], const T &t) {
    return generate(code) &&
           generate(*reinterpret_cast<const unsigned char(*)[sizeof t]>(&t));
}

const unsigned char code_en[] = {0x55, 0x8B, 0xEC}; // push ebp; mov ebp,esp
const unsigned char code_ex[] = {0x5D, 0xC3}; // pop ebp; ret
const unsigned char code_pu[] = {0x50}; // push eax
const unsigned char code_dr[] = {0x8B, 0}; // mov eax, [eax]
const unsigned char code_ld[] = {0xB8}; // mov eax, immediate
const unsigned char code_gt[] = {0xFF, 0x25}; // jmp dword ptr [id]
const unsigned char code_as[] = {0x59, 0x89, 0x01}; // pop ecx; mov [ecx], eax
const unsigned char code_aa[] = {0x8D, 0x04, 0x85}; // lea eax, [eax * 4 + ?]
const unsigned char code_ad[] = {0x59, 0x03, 0xC1}; // pop ecx; add eax, ecx
const unsigned char code_sb[] = {0x8B, 0xC8, 0x58, 0x2B, 0xC1};
                                            // mov ecx,eax; pop eax; sub eax,ecx
const unsigned char code_ml[] = {0x59, 0xF7, 0xE9}; // pop ecx; imul ecx
const unsigned char code_dv[] = {0x8B, 0xC8, 0x58, 0x99, 0xF7, 0xF9};
                                         // mov ecx, eax; pop eax; cdq; idiv ecx
const unsigned char code_md[] = {0x8B, 0xC2}; // mov eax, edx
const unsigned char code_is[] = {0xF7, 0xD8}; // neg eax

bool entry_code() {        return generate(code_en); }
bool exit_code() {         return generate(code_ex); }
bool push() {              return generate(code_pu); }
bool dereference() {       return generate(code_dr); }
bool load(void *p) {       return generate(code_ld, p); }
bool load(int i) {         return generate(code_ld, i); }
bool go_to(int *id) {      return generate(code_gt, id); }
bool assign() {            return generate(code_as); }
bool array_address(int*id){return generate(code_aa, id); }
bool add() {               return generate(code_ad); }
bool subtract() {          return generate(code_sb); }
bool multiply() {          return generate(code_ml); }
bool divide() {            return generate(code_dv); }
bool modulo(){ return divide() && generate(code_md); }
bool inverse_sign() {      return generate(code_is); }
bool unchange() {          return true; }

bool relational_operation(unsigned char jcode) {
    char unsigned code[]={0x59,0x3B,0xC1,0x0F,jcode}; // pop ecx;cmp eax,ecx;j??
    generate(code, 0);
    *ct->jump_ptr++ = ct->instruction_ptr;
    return true;
}

bool equal() {            return relational_operation(0x85); } // jne
bool not_equal() {        return relational_operation(0x84); } // je
bool less() {             return relational_operation(0x8E); } // jle
bool less_or_equal() {    return relational_operation(0x8C); } // jl
bool greater() {          return relational_operation(0x8D); } // jge
bool greater_or_equal() { return relational_operation(0x8F); } // jg

bool end_if() {
    char *p = *--ct->jump_ptr;
    reinterpret_cast<int*>(p)[-1] = ct->instruction_ptr - p;
    return true;
}

template<typename T> bool call(T function_address) {
    static const unsigned char code[]={0x8B,0xC8,0xE8}; // mov ecx, eax;call rel
    return generate(code, reinterpret_cast<char*>(function_address) -
        ct->instruction_ptr - (sizeof code) - sizeof(void*));
}

void _fastcall print_int_main(int i)      { cout <<  i << flush; }
void _fastcall print_text_main(string *s) { cout << *s << flush; }
bool print_int() {           return            call(print_int_main); }
bool print_text(string *s) { return load(s) && call(print_text_main); }

bool label(int *id) { // Warnning! Cannot undo this action by ~test().
    *id = reinterpret_cast<int>(ct->instruction_ptr);
    return true;
}

class text_eater {
    bool escaping, sjis;
    string &s;
    char mark;
public:
    text_eater(string &s, char mark) :
        escaping(false), sjis(false), s(s), mark(mark) {}
    bool operator()(unsigned char c) {
        if (sjis) {
            s.push_back(c);
            return sjis = false;
        } else if (escaping) {
            switch(c) {
                case 'a': s.push_back('\a'); break;
                case 'n': s.push_back('\n'); break;
                case 't': s.push_back('\t'); break;
                default:  s.push_back(c);
            }
            return escaping = false;
        }
        if (c == mark || c == '\n') return true;
        escaping = (c == '\\');
        if (!escaping) s.push_back(c);
        sjis = (0x80 < c && c < 0xA0 || 0xDF < c && c < 0xFD);
        return false;
    }
};

bool text_string(string &s, char mark) {
    ct->code_itr = find_if(ct->code_itr, code.end(), text_eater(s = "", mark));
    if (ct->code_itr == code.end() || *ct->code_itr != mark) return false;
    ++ct->code_itr;
    return true;
}

bool not_is_space(char c) { return !isspace(c) || c == '\n'; }
bool not_is_ident(char c) { return !isalnum(c) && c != '_'; }
bool not_is_digit(char c) { return !isdigit(c); }

bool token(string &s, bool &is_text) {
    is_text = false;
    ct->code_itr = find_if(ct->code_itr, code.end(), not_is_space);
    string::iterator it = ct->code_itr;
    char c = (it == code.end() ? 0 : *it), c3[3] = { c, 0, 0 };
    switch (c) {
        case ':': case '=': case '!': case '>': case '<': case '+': case '-':
        case '*': case '/': case '%':
            if (++it != code.end() && *it == '=') {
                ++ct->code_itr;
                c3[1] = *it;
            }
        case '\n': case ';': case '(': case ')': case '[': case ']':
            ++ct->code_itr;
        case 0:
            s = c3;
            return c != 0;
        case '\"':
            ++ct->code_itr;
            return is_text = text_string(s, c);
    }
    if (!isalnum(*it)) return false;
    bool (*not_is)(char) = (isdigit(*it) ? not_is_digit : not_is_ident);
    string(it, ct->code_itr = find_if(it, code.end(), not_is)).swap(s);
    return true;
}

bool is(const string &s) {
    string sym;
    bool is_text;
    return test() && token(sym, is_text) && !is_text && sym == s && ok();
}

bool text(string *&s) {
    bool is_text;
    return test() && (s=ct->text_ptr++), token(*s, is_text) && is_text && ok();
}

bool relational_operator(bool (*&op)()) {
    return is("=") && (op = equal) || is("!=") && (op = not_equal) ||
        is("<") && (op = less) || is("<=") && (op = less_or_equal) ||
        is(">") && (op = greater) ||is(">=") && (op = greater_or_equal);
}

bool (*unary_operator())() {
    return is("-") ? inverse_sign : (is("+"), unchange);
}

bool multiplicative_operator(bool (*&op)()) {
    return is("*") && (op = multiply) || is("/") && (op = divide) ||
        is("%") && (op = modulo);
}

bool additive_operator(bool (*&op)()) {
    return is("+") && (op = add) || is("-") && (op = subtract);
}

bool create_identifier(const string &s, int *&id) {
    ct->identifiers[s] = id = ct->data_ptr++;
    return true;
}

bool identifier(string &s) {
    bool is_text;
    return test() && token(s, is_text) && !is_text && isalpha(s[0]) && ok();
}

bool identifier(int *&id, bool must_be_new = false) {
    string s;
    return test() && identifier(s) && (ct->contains(s, id) ? !must_be_new :
        create_identifier(s, id)) && ok();
}

bool bracketed_expression(int *id) {
    return test() && is("[") && expression() && is("]") &&
        array_address(id) && ok();
}

bool variable_address() {
    int *id;
    return identifier(id) && (bracketed_expression(id) || load(id));
}

bool variable() { return variable_address() && dereference(); }

bool number(int &i, bool (*op)() = unchange) {
    string s;
    bool is_text;
    return test() && token(s, is_text) && !is_text && isdigit(s[0]) &&
        (i = atoi(s.c_str()), op == inverse_sign && (i = -i), ok());
}

bool parenthesised_expression() {
    return test() && is("(") && expression() && is(")") && ok();
}

bool factor(bool (*op)()) {
    int i;
    return number(i, op) && load(i) || variable() && op() ||
        parenthesised_expression() && op();
}

bool term(bool (*op)() = 0) {
    return test() && op && push(), factor(unary_operator()) &&
        (op && op(), !multiplicative_operator(op) || term(op)) && ok();
}

bool expression(bool (*op)()) {
    return test() && op && push(), term() &&
        (op && op(), !additive_operator(op) || expression(op)) && ok();
}

bool condition() {
    bool (*op)();
    return test() && expression() && push() && relational_operator(op) &&
        expression() && op() && ok();
}

bool bracketed_number(int &i) {
    return test() && is("[") && number(i) && is("]") && ok();
}

bool null_statement() { return true; }

bool print_statement() {
    string *s;
    return expression() && print_int() || text(s) && print_text(s);
}

bool assign_statement() {
    return test() && variable_address() && push() && is("=") && expression() &&
        assign() && ok();
}

bool if_statement() {
    bool statement();
    return test() && is("if") && condition() && is("then") && statement() &&
        end_if() && ok();
}

bool goto_statement() {
    int *id;
    return test() && is("goto") && identifier(id) && go_to(id) && ok();
}

bool label_definition() {
    int *id;
    return test() && identifier(id) && is(":") && label(id) && ok();
}

bool array_definition() {
    int i, *id;
    return test() && is("array") && identifier(id, true) &&
        bracketed_number(i) && (ct->data_ptr += i) && ok();
}

bool statement() {
    return array_definition() || label_definition() || goto_statement() ||
        if_statement() || assign_statement() || print_statement() ||
        null_statement();
}

bool new_line() { return is("\n") || is(";"); }
bool program() { return statement() && (!new_line() || program()); }
