//  A small programming language compiler for Visual C++ on Windows x86(32bit).
//  Full source code of the compiler included in this file.
//
//  COPYRIGHT : Public domain
//
//  HOW TO BUILD :
//    > cl /Os /O1 /Gr /GS- compiler_min_ver.c kernel32.lib
//                                 or
//    > gcc -Os -s -nostdlib -nodefaultlibs -fno-builtin compiler_min_ver.c
//         -Wl,--entry,_mainCRTStartup,--enable-stdcall-fixup -lkernel32
//
//      # Cannot use -fwhole-program option, because of the optimizer bug.
//
//  USAGE : Run it in console, it reads the program text from standard-input,
//      compiles and executes on-memory. No command line parameters. When
//      inputting from keyboard, don't forget to input CONTROL-Z to end input.
//
//  GRAMMAR : Written in BNF below.
//
//
//  program ::= statement [new_line program]
//
//  statement ::=
//      array_definition | label_definition | goto_statement | if_statement |
//      assign_statement | print_statement | null_statement
//
//  array_definition ::= 'array' identifier bracketed_number
//
//  label_definition ::= identifier ':'
//
//  if_statement ::= 'if' condition 'then' statement
//
//  goto_statement ::= 'goto' identifier
//
//  assign_statement ::= variable '=' expression
//
//  print_statement ::= expression | text
//
//  null_statement ::=
//
//  condition ::= expression relational_operator expression
//
//  expression ::= term [additive_operator expression]
//
//  term ::= [unary_operator] factor [multiplicative_operator term]
//
//  factor ::= number | variable | parenthesised_expression
//
//  parenthesised_expression ::= '(' expression ')'
//
//  bracketed_expression ::= '[' expression ']'
//
//  bracketed_number ::= '[' number ']'
//
//  variable ::= identifier [bracketed_expression]
//
//  identifier ::= alphabet [identifier_subsequence]
//
//  identifier_subsequence ::= identifier_letter [identifier_subsequence]
//
//  number ::= digit [number]
//
//  text ::= '"' text_letters '"'
//
//  new_line ::= '\n' | ';'
//
//  additive_operator ::= '+' | '-'
//
//  multiplicative_operator ::= '*' | '/' | '%'
//
//  unary_operator ::= '+' | '-'
//
//  relational_operator ::= '=' | '!=' | '<' | '<=' | '>' | '>='
//
//  identifier_letter ::= digit | alphabet | '_'
//
//  digit ::= '0' .. '9'
//
//  alphabet ::= 'A' .. 'Z' | 'a' .. 'z'
//
//  text_letters ::= any 0 or more characters, but '\' '"' sequence represents a
//      quotation mark, '\' 'a' is a bell, '\' 'n' is a line feed, '\' 't' is a
//      horizontal tab and '\' '\' sequence is a backslash itself.

#include <windows.h>
#pragma comment(linker, "/subsystem:console")
#pragma comment(linker, "/nodefaultlib")
#pragma comment(linker, "/entry:\"mainCRTStartup\"")
#pragma comment(linker, "/merge:.rdata=.text")

typedef struct dictionary dictionary;
struct dictionary {
    dictionary *less, *greater;
    char *first;
    int *second;
};

typedef struct test_struct test_struct;
struct test_struct {
    test_struct *parent;
    dictionary *identifiers;
    int *data_ptr;
    char *instruction_ptr, **jump_ptr, **text_ptr, *code_ptr;
};

int data[4096*4096];
char *instructions, *jump_stack[4096], *texts[4096], code[4096*4096], *code_end;
test_struct *ct; // 'ct' is short for 'current test_struct'.
int (_cdecl*printf_)(const char*, ...), (_cdecl*atoi_)(const char*);
int (_cdecl*isdigit_)(int), (_cdecl*isspace_)(int), (_cdecl*isalpha_)(int);
int (_cdecl*isalnum_)(int);
void *stdin_, *(_cdecl*calloc_)(size_t,size_t), (_cdecl*free_)(void*);
void *(_cdecl*memcpy_)(void*, const void*, size_t);
char *(_cdecl*strdup_)(const char*), *(_cdecl*fgets_)(char*,int,void*);
unsigned long(_stdcall*timeGetTime_)(void);

int program(void), statement(void), expression(void);
int entry_code(void), exit_code(void);

void load_libraries(void) {
    void *h = LoadLibraryA("msvcrt.dll");
    printf_ = (int(_cdecl*)(const char*, ...))GetProcAddress(h, "printf");
    atoi_ = (int(_cdecl*)(const char*))GetProcAddress(h, "atoi");
    isdigit_ = (int(_cdecl*)(int))GetProcAddress(h, "isdigit");
    isspace_ = (int(_cdecl*)(int))GetProcAddress(h, "isspace");
    isalpha_ = (int(_cdecl*)(int))GetProcAddress(h, "isalpha");
    isalnum_ = (int(_cdecl*)(int))GetProcAddress(h, "isalnum");
    stdin_ = (void*)GetProcAddress(h, "_iob");
    calloc_ = (void*(_cdecl*)(size_t,size_t))GetProcAddress(h, "calloc");
    free_ = (void(_cdecl*)(void*))GetProcAddress(h, "free");
    memcpy_ = (void*(_cdecl*)(void*, const void*, size_t))GetProcAddress(
        h, "memcpy");
    strdup_ = (char*(_cdecl*)(const char*))GetProcAddress(h, "_strdup");
    fgets_ = (char*(_cdecl*)(char*,int,void*))GetProcAddress(h, "fgets");
    timeGetTime_ = (unsigned long(_stdcall*)(void))GetProcAddress(
        LoadLibraryA("winmm.dll"), "timeGetTime");
}

dictionary *merge(dictionary **root, dictionary *added_one) {
    if (!added_one) return *root;
    if (*root) {
        dictionary tmp = { 0, 0, added_one->first, added_one->second};
        int cmp = lstrcmpA((**root).first, tmp.first);
        merge((cmp > 0 ? &(**root).less : &(**root).greater), &tmp);
    } else {
        *root = (dictionary*)calloc_(1, sizeof **root);
        (**root).first  = strdup_(added_one->first);
        (**root).second = added_one->second;
    }
    merge(root, added_one->less);
    return merge(root, added_one->greater);
}

void dispose(dictionary **root) {
    if (*root == 0) return;
    dispose(&(**root).less);
    dispose(&(**root).greater);
    free_((**root).first);
    free_(*root);
    *root = 0;
}

dictionary *find(dictionary *root, const char *s) {
    int cmp = root ? lstrcmpA(root->first, s) : 0;
    if (cmp == 0) return root;
    return find((cmp > 0 ? root->less : root->greater), s);
}

int contains(test_struct *t, char *s, int **address) {
    dictionary *found;
    if (t->parent && contains(t->parent, s, address)) return 1;
    return (found = find(t->identifiers, s)) && (*address = found->second);
}

int test(void) {
    test_struct *t = (test_struct*)calloc_(1, sizeof *t);
    *t = (test_struct){ ct, 0, ct ? ct->data_ptr : data,
        ct ? ct->instruction_ptr : instructions, ct ? ct->jump_ptr : jump_stack,
        ct ? ct->text_ptr : texts, ct ? ct->code_ptr : code };
    ct = t;
    return 1;
}

int undo(void) {
    test_struct *t = ct;
    dispose(&ct->identifiers);
    ct = ct->parent;
    free_(t);
    return 0;
}

int ok(void) {
    dictionary *cpi = merge(&ct->parent->identifiers, ct->identifiers);
    test_struct *cpp = ct->parent->parent;
    *ct->parent = *ct;
    ct->parent->parent = cpp;
    ct->parent->identifiers = cpi;
    return !undo();
}

void read_program_text(void) {
    int rest = (sizeof code) / (sizeof *code), len;
    for (code_end = code; fgets_(code_end, rest, stdin_); ) {
        code_end += len = lstrlenA(code_end);
        rest -= len;
        if (rest > 0) continue;
        printf_("Too long program.\n");
        ExitProcess(1);
    }
}

void syntax_error(void) {
    int line = 1, i;
    for (i = 0; i < ct->code_ptr - code; ++i) line += (code[i] == '\n');
    printf_("Syntax error at line %d.\n", line);
    ExitProcess(1);
}

void compile(void) {
    read_program_text();
    instructions = (char*)VirtualAlloc(0,4096*4096,MEM_COMMIT,PAGE_READWRITE);
    test(); // Prepares 'ct' : current test_struct (of compilation).
    entry_code();
    program(); // Parses and generates code.
    if (ct->code_ptr != code_end) syntax_error();
    exit_code();
    VirtualProtect(instructions, 4096*4096, PAGE_EXECUTE, (unsigned long[]){0});
    undo(); // Doesn't undo, but cleans up 'ct'.
}

void run(void) {
    int c = timeGetTime_();
    ((void(*)(void))instructions)(); // Calls compiled code.
    c = timeGetTime_() - c;
    printf_("\nCalculated in %d.%03d seconds.\n", c / 1000, c % 1000);
}

int mainCRTStartup(void) {
    load_libraries();
    compile();
    run();
    return 0;
}

int generate(const char *code, size_t size) {
    memcpy_(ct->instruction_ptr, code, size);
    ct->instruction_ptr += size;
    return 1;
}

int generate_2(const char *code, size_t size, void *p) {
    return generate(code,size) && generate((char*)&p, sizeof p);
}

const char code_en[] = {0x55, 0x8B, 0xEC}; // push ebp; mov ebp,esp
const char code_ex[] = {0x5D, 0xC3}; // pop ebp; ret
const char code_pu[] = {0x50}; // push eax
const char code_dr[] = {0x8B, 0}; // mov eax, [eax]
const char code_ld[] = {0xB8}; // mov eax, immediate
const char code_gt[] = {0xFF, 0x25}; // jmp dword ptr [id]
const char code_as[] = {0x59, 0x89, 0x01}; // pop ecx; mov [ecx], eax
const char code_aa[] = {0x8D, 0x04, 0x85}; // lea eax, [eax * 4 + ?]
const char code_ad[] = {0x59, 0x03, 0xC1}; // pop ecx; add eax, ecx
const char code_sb[] = {0x8B, 0xC8, 0x58, 0x2B, 0xC1};
                                          // mov ecx, eax; pop eax; sub eax, ecx
const char code_ml[] = {0x59, 0xF7, 0xE9}; // pop ecx; imul ecx
const char code_dv[] = {0x8B, 0xC8, 0x58, 0x99, 0xF7, 0xF9};
                                         // mov ecx, eax; pop eax; cdq; idiv ecx
const char code_md[] = {0x8B, 0xC2}; // mov eax, edx
const char code_is[] = {0xF7, 0xD8}; // neg eax

int entry_code(void) {   return             generate(code_en, sizeof code_en); }
int exit_code(void) {    return             generate(code_ex, sizeof code_ex); }
int push(void) {         return             generate(code_pu, sizeof code_pu); }
int dereference(void) {  return             generate(code_dr, sizeof code_dr); }
int load_ptr(void *p) {  return  generate_2(code_ld, sizeof code_ld, p); }
int load(int i) {        return  generate_2(code_ld, sizeof code_ld, (void*)i); }
int go_to(int *id) {     return  generate_2(code_gt, sizeof code_gt, id); }
int assign(void) {       return             generate(code_as, sizeof code_as); }
int array_address(int*id){return generate_2(code_aa, sizeof code_aa, id); }
int add(void) {          return             generate(code_ad, sizeof code_ad); }
int subtract(void) {     return             generate(code_sb, sizeof code_sb); }
int multiply(void) {     return             generate(code_ml, sizeof code_ml); }
int divide(void) {       return             generate(code_dv, sizeof code_dv); }
int modulo(void) {       return divide() && generate(code_md, sizeof code_md); }
int inverse_sign(void) { return             generate(code_is, sizeof code_is); }
int unchange(void) {     return 1; }

int relational_operation(char jcode) {
    char code[] = {0x59, 0x3B, 0xC1, 0x0F, jcode}; // pop ecx;cmp eax,ecx;j??
    generate_2(code, sizeof code, (void*)0);
    *ct->jump_ptr++ = ct->instruction_ptr;
    return 1;
}

int equal(void) {            return relational_operation(0x85); } // jne
int not_equal(void) {        return relational_operation(0x84); } // je
int less(void) {             return relational_operation(0x8E); } // jle
int less_or_equal(void) {    return relational_operation(0x8C); } // jl
int greater(void) {          return relational_operation(0x8D); } // jge
int greater_or_equal(void) { return relational_operation(0x8F); } // jg

int end_if(void) {
    char *p = *--ct->jump_ptr;
    *((int*)p - 1) = ct->instruction_ptr - p;
    return 1;
}

int call(void *function_address) {
    static const char code[] = {0x8B, 0xC8, 0xE8}; // mov ecx, eax; call rel
    return generate_2(code, sizeof code, (void*)((char*)function_address -
        ct->instruction_ptr - (sizeof code) - sizeof(void*)));
}

void _fastcall print_text_main(char **s) { printf_("%s", *s); }
void _fastcall print_int_main(int i) { printf_("%d", i); }
int print_int(void) { return call(print_int_main); }
int print_text(char **s) { return load_ptr(s) && call(print_text_main); }

int label(int *id) { // Warnning! Cannot undo this action by undo().
    *id = (int)ct->instruction_ptr;
    return 1;
}

typedef struct {
    int escaping, sjis;
    char *s, mark;
} text_eater;

int eat(text_eater *t, unsigned char c) {
    if (t->sjis) {
        *t->s++ = c;
        return t->sjis = 0;
    } else if (t->escaping) {
        switch(c) {
            case 'a': *t->s++ = '\a'; break;
            case 'n': *t->s++ = '\n'; break;
            case 't': *t->s++ = '\t'; break;
            default:  *t->s++ = c;
        }
        return t->escaping = 0;
    }
    if (c == t->mark || c == '\n') return 1;
    t->escaping = c == '\\';
    if (!t->escaping) *t->s++ = c;
    t->sjis = (0x80 < c && c < 0xA0 || 0xDF < c && c < 0xFD);
    return 0;
}

char *text_eating_find_if(char *b, char *e, text_eater *t) {
    for (; b != e; ++b) if (eat(t,*b)) break;
    return b;
}

char *find_if(char *b, char *e, int (*is_found)(char)) {
    for (; b != e; ++b) if (is_found(*b)) break;
    return b;
}

int text_string(char **s, char mark) {
    static char buf[4096*16];
    text_eater t = {0, 0, buf, mark};
    ct->code_ptr = text_eating_find_if(ct->code_ptr, code_end, &t);
    if (ct->code_ptr == code_end || *ct->code_ptr != mark) return 0;
    ++ct->code_ptr;
    *t.s = 0;
    *s = strdup_(buf);
    return 1;
}

int not_is_space(char c) { return !isspace_(c) || c == '\n'; }
int not_is_ident(char c) { return !isalnum_(c) && c != '_'; }
int not_is_digit(char c) { return !isdigit_(c); }

int token(char **s, int *is_text) {
    int (*not_is)(char) = not_is_ident;
    char *p = ct->code_ptr = find_if(ct->code_ptr, code_end, not_is_space);
    char c = (p == code_end ? 0 : *p), c3[] = { c, 0, 0 };
    *is_text = 0;
    switch (c) {
        case ':': case '=': case '!': case '>': case '<': case '+': case '-':
        case '*': case '/': case '%':
            if (++p != code_end && *p == '=') {
                ++ct->code_ptr;
                c3[1] = *p;
            }
        case '\n': case ';': case '(': case ')': case '[': case ']':
            ++ct->code_ptr;
            *s = strdup_(c3);
        case 0:
            return c != 0;
        case '\"':
            ++ct->code_ptr;
            return *is_text = text_string(s, c);
    }
    if (!isalnum_(*p)) return 0;
    if (isdigit_(*p)) not_is = not_is_digit;
    ct->code_ptr = find_if(p, code_end, not_is);
    *s = (char*)memcpy_(calloc_(ct->code_ptr - p + 1, 1), p, ct->code_ptr - p);
    return 1;
}

int is(char *s) {
    char *sym = 0;
    int is_text, result = test() && token(&sym, &is_text) && !is_text &&
        lstrcmpA(sym,s) == 0 && ok() || undo();
    free_(sym);
    return result;
}

int text(char ***s) {
    int is_text, result = test() && token(*s = ct->text_ptr++, &is_text) &&
                    is_text && ok() || undo();
    if (result) return 1;
    free_(**s);
    **s = 0;
    return 0;
}

int relational_operator(int (**op)(void)) {
    return is("=") && (*op = equal) || is("!=") && (*op = not_equal) ||
        is("<") && (*op = less) || is("<=") && (*op = less_or_equal) ||
        is(">") && (*op=greater) || is(">=") && (*op = greater_or_equal);
}

int (*unary_operator(void))(void) {
    return is("-") ? inverse_sign : (is("+"), unchange);
}

int multiplicative_operator(int (**op)(void)) {
    return is("*") && (*op = multiply) || is("/") && (*op = divide) ||
        is("%") && (*op = modulo);
}

int additive_operator(int (**op)(void)) {
    return is("+") && (*op = add) || is("-") && (*op = subtract);
}

int create_identifier(char *s, int **id) {
    merge(&ct->identifiers, &(dictionary){ 0, 0, s, *id = ct->data_ptr++ });
    return 1;
}

int identifier_str(char **s) {
    int is_text;
    return test() && token(s, &is_text) && !is_text && isalpha_(**s) &&
        ok() || undo();
}

int identifier_cond(int **id, int must_be_new) {
    char *s = 0;
    int result = test() && identifier_str(&s) && (contains(ct, s, id) ?
        !must_be_new : create_identifier(s, id)) && ok() || undo();
    free_(s);
    return result;
}

int identifier(int **id) { return identifier_cond(id, 0); }

int bracketed_expression(int *id) {
    return test() && is("[") && expression() && is("]") && array_address(id) &&
        ok() || undo();
}

int variable_address(void) {
    int *id;
    return identifier(&id) && (bracketed_expression(id) || load_ptr(id));
}

int variable(void) { return variable_address() && dereference(); }

int number_op(int *i, int (*op)(void)) {
    char *s = 0;
    int is_text, result = test() && token(&s, &is_text) && !is_text &&
        isdigit_(s[0]) &&
        (*i = atoi_(s), op == inverse_sign && (*i = -*i), ok()) || undo();
    free_(s);
    return result;
}

int number(int *i) { return number_op(i, unchange); }

int parenthesised_expression(void) {
    return test() && is("(") && expression() && is(")") && ok() || undo();
}

int factor(int (*op)(void)) {
    int i;
    return number_op(&i, op) && load(i) || variable() && op() ||
        parenthesised_expression() && op();
}

int term_op(int (*op)(void)) {
    return test() && op && push(), factor(unary_operator()) && (op && op(),
        !multiplicative_operator(&op) || term_op(op)) && ok() || undo();
}

int term(void) { return term_op(0); }

int expression_op(int (*op)(void)) {
    return test() && op && push(), term() && (op && op(),
        !additive_operator(&op) || expression_op(op)) && ok() || undo();
}

int expression(void) { return expression_op(0); }

int condition(void) {
    int (*op)(void);
    return test() && expression() && push() && relational_operator(&op) &&
        expression() && op() && ok() || undo();
}

int bracketed_number(int *i) {
    return test() && is("[") && number(i) && is("]") && ok() || undo();
}

int null_statement(void) { return 1; }

int print_statement(void) {
    char **s;
    return expression() && print_int() || text(&s) && print_text(s);
}

int assign_statement(void) {
    return test() && variable_address() && push() && is("=") && expression() &&
        assign() && ok() || undo();
}

int if_statement(void) {
    return test() && is("if") && condition() && is("then") && statement() &&
        end_if() && ok() || undo();
}

int goto_statement(void) {
    int *id;
    return test() && is("goto") && identifier(&id) && go_to(id) && ok()||undo();
}

int label_definition(void) {
    int *id;
    return test() && identifier(&id) && is(":") && label(id) && ok() || undo();
}

int array_definition(void) {
    int i, *id;
    return test() && is("array") && identifier_cond(&id, 1) &&
        bracketed_number(&i) && (ct->data_ptr += i) && ok() || undo();
}

int statement(void) {
    return array_definition() || label_definition() || goto_statement() ||
        if_statement() || assign_statement() || print_statement() ||
        null_statement();
}

int new_line(void) { return is("\n") || is(";"); }
int program(void) { return statement() && (!new_line() || program()); }
